#pragma once

#include <algorithm>

struct SlewLimiter {
  float out = 0.f;
  float rise = 0.f;
  float fall = 0.f;

  void reset() {
    out = 0.f;
  }

  void setRiseFall(float rise, float fall) {
    this->rise = rise;
    this->fall = fall;
  }
  float process(float deltaTime, float in) {
    out = clamp(in, out - fall * deltaTime, out + rise * deltaTime);
    return out;
  }
  float clamp(float x, float a, float b) {
    return std::max(std::min(x, b), a);
  }
};

class AttackSustainRelease {
public:
  enum Stage {
    OPEN,
    CLOSING,
    CLOSED
  };
  SlewLimiter sl;
  AttackSustainRelease();
  void reset();
  void release();
  float process(float deltaTime);
  float getOutput();
  Stage getStage();
  bool isOpen();
  bool isClosing();
  bool isClosed();
  void setAttackTime(float attackTime);
  void setReleaseTime(float releaseTime);
  void setAttackReleaseTime(float attackTime, float releaseTime);

private:
  Stage stage;
  bool open;
  bool closing;
  float out;
  float target;
  float releaseRate = 1.f;
  float attackRate = 1.f;
};

void AttackSustainRelease::reset() {
  stage = OPEN;
  open = false;
  closing = false;
  out = 0.f;
  target = 1.f;
  sl.reset();
}

AttackSustainRelease::AttackSustainRelease() {
  reset();
}

void AttackSustainRelease::setAttackReleaseTime(float attackTime, float releaseTime) {
  sl.setRiseFall(1.f / attackTime, 1.f / releaseTime);
}

void AttackSustainRelease::release() {
  stage = CLOSING;
  target = 0.f;
}

float AttackSustainRelease::process(float deltaTime) {
  out = sl.process(deltaTime, target);

  // Close envelope after fade to zero
  if (stage == CLOSING && out == 0.f) {
    stage = CLOSED;
  }
  return out;
}

float AttackSustainRelease::getOutput() {
  return out;
}

AttackSustainRelease::Stage AttackSustainRelease::getStage() {
  return stage;
}

bool AttackSustainRelease::isOpen() {
  return stage == AttackSustainRelease::OPEN;
};

bool AttackSustainRelease::isClosing() {
  return stage == AttackSustainRelease::CLOSING;
}

bool AttackSustainRelease::isClosed() {
  return stage == AttackSustainRelease::CLOSED;
}
